"""
This client uses a channel to send and receive possibly faulty messages
CrySyS Lab
2019
"""

import socket
import string
import random
import atexit
import logging

# Task: import the required modules
from PyCRC.CRC16 import CRC16

import unireedsolomon
# End Task

import message_format
import server_address

DEFAULT_ECC_TYPE = None

# TODO Uncomment it in Task 1
# DEFAULT_ECC_TYPE = message_format.ECC_TYPE_NONE

# TODO Uncomment it in Task 2
# DEFAULT_ECC_TYPE = message_format.ECC_TYPE_CRC16

# TODO Uncomment it in Task 3 and 4
# DEFAULT_ECC_TYPE = message_format.ECC_TYPE_RS5

DEFAULT_PAYLOAD_LENGTH = 20

TEST_MESSAGE = ''
TODO = ''


def create_random_payload(length):
    """Creates a random payload with the given length to be transferred (or TEST_MESSAGE if it is defined)"""

    if TEST_MESSAGE:
        return TEST_MESSAGE

    if length <= 0 or message_format.PAYLOAD_MAX_LENGTH < length:
        return ''

    letters = string.ascii_lowercase
    string_length = length
    payload = ''.join(random.choice(letters) for _ in range(string_length))
    logging.debug("Random payload generated: " + payload)
    return payload


def create_ecc(message, ecc_type):
    """Creates the given type ECC for the given message """

    ecc_length = 0
    ecc_value = b''

    if message_format.ECC_TYPE_NONE == ecc_type:
        pass

    elif message_format.ECC_TYPE_CRC16 == ecc_type:
        # Task 2: Create CRC16
        # Hint: Check how the received code is checked
        ecc_value_int = int(TODO)
        print("ECC value:", ecc_value_int)
        ecc_value = ecc_value_int.to_bytes(message_format.ECC_LENGTH_CRC16, byteorder="big")
        ecc_length = TODO
        # End task

    elif message_format.ECC_TYPE_RS5 == ecc_type:
        # Task 3: Create RS code that can correct up to 5 errors
        # Hint: Check how the received code is checked
        encoder = TODO
        ecc_value = encoder.encode(message, return_string=False)
        logging.debug("Symbol generated: " + repr(ecc_value))
        # Hint: the ECC value is located after the message
        ecc_value = TODO
        tmp = b''
        for value in ecc_value:
            # Hint: before concatenation each int value should be converted to bytes using the following:
            #  integer_value.to_bytes(number_of_bytes, byteorder="big")
            tmp = TODO
        ecc_value = tmp
        logging.debug("ECC value: " + str(ecc_value))
        ecc_length = TODO
        # End task

    else:
        logging.critical("Unknown ecc type")
        raise ValueError

    return ecc_type, ecc_length, ecc_value


def sender_role(sock, ecc_type):
    """Sends random messages with the given ECC type"""

    logging.info("Sender role selected")
    while True:
        payload_length = DEFAULT_PAYLOAD_LENGTH

        payload_content = create_random_payload(payload_length)
        try:
            ecc_type, ecc_length, ecc_value = create_ecc(payload_content, ecc_type)
        except ValueError:
            logging.critical("Cannot calculate ecc, exiting")
            break

        message = message_format.TYPE_TRANSFER.to_bytes(1, byteorder="big") + \
            payload_length.to_bytes(1, byteorder="big") + \
            bytearray(str(payload_content).encode()) + \
            ecc_type.to_bytes(1, byteorder="big") + \
            ecc_length.to_bytes(1, byteorder="big") + \
            ecc_value

        logging.debug("Sending message: " + str(message))
        sock.sendto(message, (server_address.UDP_IP, server_address.UDP_PORT))

        input("Hit enter to send next message")


def receiver_role(sock):
    """Receives messages and checks the ECC"""

    logging.info("Receiver role selected")
    while True:
        data, address = sock.recvfrom(server_address.UDP_MAX_BUF_SIZE)
        logging.debug("Received message: " + str(data) + " from: " + str(address))

        message_type = data[0]
        message_length = data[message_format.LENGTH_MESSAGE_TYPE]
        message = data[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH:
                       message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + message_length]
        ecc_type = data[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + message_length]
        ecc_length = data[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + message_length +
                          message_format.LENGTH_ECC_TYPE]
        ecc_value = data[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + message_length +
                         message_format.LENGTH_ECC_TYPE + message_format.LENGTH_ECC_LENGTH:
                         message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + message_length +
                         message_format.LENGTH_ECC_TYPE + message_format.LENGTH_ECC_LENGTH+ecc_length]

        logging.debug(str(message_type) + " " + str(message_length) + " " + str(message) + " " +
                      str(ecc_type) + " " + str(ecc_length) + " " + str(ecc_value))

        if message_format.ECC_TYPE_NONE == ecc_type:
            print("No ECC defined, nothing to check")

        elif message_format.ECC_TYPE_CRC16 == ecc_type:
            # Check the CRC16 value and print whether it was correct
            calculated_ecc_value = CRC16().calculate(message)
            print("ECC:", calculated_ecc_value)
            calculated_ecc_value = calculated_ecc_value.to_bytes(message_format.ECC_LENGTH_CRC16, byteorder="big")
            if ecc_value == calculated_ecc_value:
                print("Good CRC16 checksum")
            else:
                print("Bad CRC16 checksum")

        elif message_format.ECC_TYPE_RS5 == ecc_type:
            encoder = unireedsolomon.rs.RSCoder(len(message) + message_format.ECC_LENGTH_RS5, len(message))
            symbol = message+ecc_value
            logging.debug("Symbol received: " + repr(symbol))
            try:
                corrected_symbol = encoder.decode(symbol)
            except unireedsolomon.rs.RSCodecError:
                print("Cannot decode due to too many errors")
                continue
            corrected_message = corrected_symbol[0].encode()
            print(repr(message), repr(corrected_message))
            if corrected_message == message:
                print("The message was correct:", corrected_message)
            else:
                print("The message was incorrect, corrected to:", corrected_message)

        else:
            logging.error("Unknown ECC type, cannot do anything")


def goodbye(sock):
    """This functions runs whenever the program exists"""

    logging.warning("Saying goodbye")
    sock.sendto(message_format.TYPE_DISCONNECT.to_bytes(1, byteorder="big"),
                (server_address.UDP_IP, server_address.UDP_PORT))


def main():
    """Main functionality"""

    # Configure logging
    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)

    logging.warning("Connecting to channel")
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # Internet, UDP
    sock.sendto(message_format.TYPE_CONNECT.to_bytes(1, byteorder="big"),
                (server_address.UDP_IP, server_address.UDP_PORT))
    atexit.register(goodbye, sock)

    role = input("Select your role, sender or receiver: s|r: ")
    if role == 's':
        sender_role(sock, DEFAULT_ECC_TYPE)
    elif role == 'r':
        receiver_role(sock)
    else:
        logging.critical("Unknown role, exiting")


main()
